"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const babel_core_1 = require("babel-core");
const expressionStringToOperatorName = {
    '+': 'add',
    '-': 'subtract',
    '*': 'multiply',
    '**': 'pow',
    '/': 'divide',
    '%': 'mod',
    '<': 'lessThan',
    '>': 'greaterThan',
    '<=': 'lessEqual',
    '>=': 'greaterEqual',
    '==': 'unsafeEqual',
    '===': 'strictEqual',
    '!=': 'unsafeNotEqual',
    '!==': 'strictNotEqual',
    '&': 'logicalAnd',
    '|': 'logicalOr',
    '^': 'logicalXor',
    '<<': 'leftShift',
    '>>': 'rightShift'
};
const unaryOperatorStringToOperatorName = {
    '~': 'logicalNot',
    '-': 'negate',
    '+': 'positive',
    '!': 'not'
};
const updateOperatorStringToOperatorName = {
    '--': 'decrement',
    '++': 'increment'
};
function isSideOnlyNumericLiterals(expression) {
    return babel_core_1.types.isNumericLiteral(expression) || (babel_core_1.types.isBinaryExpression(expression) && isBinaryExpressionOnlyNumericLiterals(expression));
}
function isBinaryExpressionOnlyNumericLiterals(node) {
    return isSideOnlyNumericLiterals(node.left) && isSideOnlyNumericLiterals(node.right);
}
function shouldBinaryExpressionBeReplaced(node) {
    if (!isBinaryOperatorSupported(node.operator)) {
        return false;
    }
    return isBinaryExpressionOnlyNumericLiterals(node);
}
function isFunctionCallUselessForAssignmentExpression(path) {
    return path.node.operator === '=';
}
function shouldUseOperatorForArgument(expression) {
    return babel_core_1.types.isMemberExpression(expression) || babel_core_1.types.isIdentifier(expression);
}
function isOperatorSupported(sourceObject, expressionString) {
    return sourceObject.hasOwnProperty(expressionString);
}
function isBinaryOperatorSupported(expressionString) {
    return isOperatorSupported(expressionStringToOperatorName, expressionString);
}
function isUnaryOperatorSupported(expressionString) {
    return isOperatorSupported(unaryOperatorStringToOperatorName, expressionString);
}
function isUpdateOperatorSupported(expressionString) {
    return isOperatorSupported(updateOperatorStringToOperatorName, expressionString);
}
function getBaseOperatorFromAssignmentExpression(assignmentOperator) {
    return assignmentOperator.substr(0, assignmentOperator.length - 1);
}
function isLValueSupported(lvalue) {
    return babel_core_1.types.isIdentifier(lvalue) || babel_core_1.types.isMemberExpression(lvalue);
}
function createOperatorCall(operatorName, ...args) {
    return babel_core_1.types.callExpression(babel_core_1.types.memberExpression(babel_core_1.types.identifier('Operator'), babel_core_1.types.identifier(operatorName)), args);
}
const requireStatement = `const {Operator} = require('macaroni');`;
const importStatement = `import {Operator} from 'macaroni'`;
const visitor = {
    BinaryExpression(path, state) {
        if (shouldBinaryExpressionBeReplaced(path.node)) {
            return;
        }
        if (!isBinaryOperatorSupported(path.node.operator)) {
            return;
        }
        const operatorName = expressionStringToOperatorName[path.node.operator];
        path.replaceWith(createOperatorCall(operatorName, path.node.left, path.node.right));
        state.hasTransformedAny = true;
    },
    AssignmentExpression(path, state) {
        if (isFunctionCallUselessForAssignmentExpression(path) || !isLValueSupported(path.node.left)) {
            return;
        }
        const expressionString = getBaseOperatorFromAssignmentExpression(path.node.operator);
        if (!isBinaryOperatorSupported(expressionString)) {
            return;
        }
        const operatorName = expressionStringToOperatorName[expressionString];
        const right = path.node.right;
        path.node.operator = '=';
        path.node.right = createOperatorCall(operatorName, path.node.left, right);
        state.hasTransformedAny = true;
    },
    UnaryExpression(path, state) {
        if (!shouldUseOperatorForArgument(path.node.argument)) {
            return;
        }
        if (!isUnaryOperatorSupported(path.node.operator)) {
            return;
        }
        const operatorName = unaryOperatorStringToOperatorName[path.node.operator];
        path.replaceWith(createOperatorCall(operatorName, path.node.argument));
        state.hasTransformedAny = true;
    },
    UpdateExpression(path, state) {
        if (!isUpdateOperatorSupported(path.node.operator)) {
            return; // should never happen unless they add a new operator. Just to be safe though.
        }
        if (!shouldUseOperatorForArgument(path.node.argument)) {
            return;
        }
        const operatorName = updateOperatorStringToOperatorName[path.node.operator];
        const variableAssignment = babel_core_1.types.assignmentExpression('=', path.node.argument, createOperatorCall(operatorName, path.node.argument));
        state.hasTransformedAny = true;
        // Yay, it's a prefix operator! Everything is easy for this one. Just replace with the call.
        if (path.node.prefix) {
            path.replaceWith(variableAssignment);
            return;
        }
        // For a postfix, we try to get this goal:
        // (temp = a && (a = Operator.increment(a) || temp) && temp))
        const tempVariable = path.scope.generateUidIdentifier('temp');
        // Hoist the variable to the scope without defining it yet
        path.scope.push({ id: tempVariable });
        const tempAssignment = babel_core_1.types.assignmentExpression('=', tempVariable, path.node.argument);
        const assignmentOrTemp = babel_core_1.types.logicalExpression('||', variableAssignment, tempVariable);
        const getPostfixValueLogical = babel_core_1.types.logicalExpression('&&', assignmentOrTemp, tempVariable);
        const assignTempAndGetValue = babel_core_1.types.logicalExpression('&&', tempAssignment, getPostfixValueLogical);
        path.replaceWith(assignTempAndGetValue);
    },
    Program: {
        enter(_, state) {
            state.hasTransformedAny = false;
            state.opts = { importType: 'import', ...(state.opts || {}) };
        },
        exit(path, state) {
            if (!state.hasTransformedAny || state.opts.skipImport) {
                return;
            }
            const importTemplate = babel_core_1.template(state.opts.importType === 'import' ? importStatement : requireStatement, { sourceType: 'module' });
            const lastImportDeclaration = path.get('body').filter(p => p.isImportDeclaration()).pop();
            if (lastImportDeclaration) {
                lastImportDeclaration.insertAfter(importTemplate());
            }
            else {
                // @ts-ignore - it does exist mr typescript
                path.unshiftContainer('body', [importTemplate()]);
            }
        }
    }
};
const transformer = () => ({ visitor });
exports.default = transformer;
//# sourceMappingURL=index.js.map